

# need source("Analysis_rfd_nex2treestring.r")
# and source("Analysis_rfd_nodetable2trlb.r")


#       Jargon

# rfd = (generalised) Robinson-Foulds distance.

# A 'trlbi' is a simple representation of a tree containing enough 
# information for rfd calculation. It consists of 4 arrays giving 
# tip names, right descendants, left descendants, branch lengths
# of the nodes in the tree. 




# parses Newick string. Finds indices of opening parenthesis, comma separating
# subtrees at root, and closing parenthesis.
find.openp.comma.closep <- function(nw) {
paren <- 0
comma <- -1
closep <- -1
openp <- -1
for (i in 1:length(nw)) {
  if (nw[i] == "(") { paren <- paren + 1 }
  if (nw[i] == ")") { paren <- paren - 1 }
  if (paren==1  &&  nw[i] == ",") {
    comma <- i
    }
  if (nw[i] == "(" &&  paren==1) {
    openp <- i
    }
  if (nw[i] == ")" &&  paren==0) {
    closep <- i
    }
  }
if (!((comma > 0  &&  closep > 0  &&  openp > 0)  ||  (comma < 0  &&  closep < 0  &&  openp < 0))) {
  cat("error\n")
  print(nw)
  }
stopifnot((comma > 0  &&  closep > 0  &&  openp > 0)  ||  (comma < 0  &&  closep < 0  &&  openp < 0))
list(openp=openp, comma=comma, closep=closep)
}


# Finds the branch length at the end of a Newick substring.
# It works by finding the last colon, and converting the text after to a number.
# It returns the text before the last colon and the branch length
extract.trlen <- function(nw) {
b <- length(nw)
stopifnot(any(grepl(":", nw)))
while (nw[b] != ":") { b <- b-1 }
trlen <- as.double(paste(nw[(b+1):length(nw)], sep="", collapse=""))
list(nw=nw[1:(b-1)], trlen=trlen)
}




# Recursively converts a Newick substring as a character vector into
# a trlbi.
subnewick.chvect.to.trlbi <- function(nw, trlbi) {
cps <- find.openp.comma.closep(nw)
openp <- cps$openp
comma <- cps$comma
closep <- cps$closep

nextni <- trlbi$nextni
nwlen <- extract.trlen(nw)
brlen <- nwlen$trlen
if (comma <= 0) {
  tipname <- paste(nwlen$nw, sep="", collapse="")
  lft <- -1
  rgt <- -1
  } else {
  tipname <- ""
  subtree.lft <- nw[2:(comma-1)]
  subtree.rgt <- nw[(comma+1):(closep-1)]
  
  trlbi <- subnewick.chvect.to.trlbi(subtree.lft, trlbi)
  nextni <- trlbi$nextni
  lft <- nextni - 1
  trlbi <- subnewick.chvect.to.trlbi(subtree.rgt, trlbi)
  nextni <- trlbi$nextni
  rgt <- nextni - 1
  }
nextni <- nextni + 1
tipnames <- c(trlbi$tipnames, tipname)
lfts <- c(trlbi$lfts, lft)
rgts <- c(trlbi$rgts, rgt)
brlens <- c(trlbi$brlens, brlen)
list(tipnames=tipnames, lfts=lfts, rgts=rgts, brlens=brlens, nextni=nextni)
}



# Converts a Newick string as into a trlbi.
newick.string.to.trlbi <- function(newick) {
newick.cv <- unlist(strsplit(newick, ""))
len <- length(newick.cv)
stopifnot(identical(newick.cv[len], ";"))
# replace final ; with :0 so root has a branch length like other nodes
newick.cv[len] <- ":"
newick.cv <- c(newick.cv, "0")
empty.trlbi <- list(tipnames=character(0), lfts=integer(0), rgts=integer(0),
                                  brlens=double(0), nextni=1)
trlbi <- subnewick.chvect.to.trlbi(newick.cv, empty.trlbi)
trlbi
}


# Removes characters representing sequences (eg "A" and "B", or "0" and "1")
# from the tipnames in a trlbi.
remove.seqchars <- function(trlbi, A, B) {
n <- length(trlbi$tipnames)
for (i in 1:n) {
  tn <- trlbi$tipnames[i]
  lastchar <- substr(tn, nchar(tn), nchar(tn))
  if (identical(lastchar, A)  ||  identical(lastchar, B)) {
    trlbi$tipnames[i] <- substr(tn, 1, nchar(tn)-1)
    }
  }
trlbi
}


# Finds the indices of tetraploid pairs in the tipnames of a trlbi.
find.tet.pairs <- function(trlbi) {
tipnames <- trlbi$tipnames
n <- length(tipnames)
tet1s <- integer(0)
tet2s <- integer(0)
for (i in 1:n) {
  tni <- tipnames[i]
  if (nchar(tni) > 0) {
    for (j in (i+1):n) {
      tnj <- tipnames[j]
      if (identical(tni, tnj)) {
        tet1s <- c(tet1s, i)
        tet2s <- c(tet2s, j)
        }
      }
    }
  }
trlbi <- list(tet1s=tet1s, tet2s=tet2s)
}


# Adds characters representing sequences (eg "A" and "B", or "0" and "1")
# to tipnames in a trlbi, using the indices in 
# This renames the tips according to the info in tetpairs. 
add.seqchars <- function(trlbi, tetpairs, A, B) {
n <- length(tetpairs$tet1s)
for (i in 1:n) {
  i1 <- tetpairs$tet1s[i]
  i2 <- tetpairs$tet2s[i]
  trlbi$tipnames[i1] <- paste(trlbi$tipnames[i1], A, sep="")  
  trlbi$tipnames[i2] <- paste(trlbi$tipnames[i2], B, sep="")
  }
trlbi
}


# Converts a Newick string with possibly-nonstandard naming of sequences
# into a trlbi with "A"s and "B"s.
newick.string.to.AB.trlbi <- function(newick, A, B) {
trlbi <- newick.string.to.trlbi(newick)
trlbi <- remove.seqchars(trlbi, A, B)
tetpairs <- find.tet.pairs(trlbi)
tetpairs
trlbi <- add.seqchars(trlbi, tetpairs, "A", "B")
trlbi
}

# Converts a Newick string into a trlbi with no sequence characters in tipnames
# plus indices of tetpairs saying where they can be added.
newick.string.to.trlbi.and.tetpairs <- function(newick, A, B) {
trlbi <- newick.string.to.trlbi(newick)
trlbi <- remove.seqchars(trlbi, A, B)
tetpairs <- find.tet.pairs(trlbi)
list(tetpairs=tetpairs, trlbi=trlbi)
}


# Recursively converts a trlbi into a list of clades and branch lengths.
# The clades are strings with comma-separated tip names like "a,xA" or "b,zB,yB".
trlbi.sub.to.clades <- function(trlbi, ni, cladelist) {
if (trlbi$lfts[ni] <= 0) {
  } else {
  lft <- trlbi$lfts[ni]
  rgt <- trlbi$rgts[ni]
  trlbicl <- trlbi.sub.to.clades(trlbi, lft, cladelist)
  trlbi <- trlbicl$trlbi
  cladelist <- trlbicl$cladelist
  trlbicl <- trlbi.sub.to.clades(trlbi, rgt, cladelist)
  trlbi <- trlbicl$trlbi
  cladelist <- trlbicl$cladelist
  trlbi$tipnames[ni] <- paste(trlbi$tipnames[lft], trlbi$tipnames[rgt], sep=",")
  }
cladelist$clades <- c(cladelist$clades, trlbi$tipnames[ni])
cladelist$brlens <- c(cladelist$brlens, trlbi$brlens[ni])
list(trlbi=trlbi, cladelist=cladelist)
}


# Converts a trlbi into a list of clades and branch lengths.
# Note that it lists all clades, not just tips.
trlbi.to.clades <- function(trlbi) {
empty.cladelist <- list(clades=character(0), brlens=double(0))
cladelist <- trlbi.sub.to.clades(trlbi, trlbi$nextni-1, empty.cladelist)$cladelist
for (i in 1:length(cladelist$clades)) {
  tips <- str_split(cladelist$clades[i], ",")[[1]]
  tips <- sort(tips)
  cladelist$clades[i] <- paste(tips, sep="", collapse=",")
  }
cladelist
}

# Converts a trlbi without sequence info into clades with
# "A"s and "B"s added according to tetpairs.         
trlbi.tetpairs.to.clades <- function(trlbi, tetpairs) {
trlbi <- add.seqchars(trlbi, tetpairs, "A", "B")
trlbi.to.clades(trlbi) 
}


# Robinson-Foulds distance between two binary rooted trees represented as clades.
# It returns rfdmiss which is the distance from clades which do not have matches
# in the other tree and rfd which is the total. Note that rfdmiss==0 iff
# there are no topological errors.
robinsonfouldsdist <- function(cb1, cb2) {
len1 <- length(cb1$clades)
len2 <- length(cb2$clades)
stopifnot(len1 == len2)

ord1 <- order(cb1$clades)
ord2 <- order(cb2$clades)
rfdhit <- 0
rfdmiss <- 0
rfdtop <- 0
i <- 1
j <- 1
while (i <= len1 && j <= len2) {
  ii <- ord1[i]
  jj <- ord2[j]
  if (identical(cb1$clades[ii], cb2$clades[jj])) {
    rfdhit <- rfdhit + abs(cb1$brlens[ii] - cb2$brlens[jj])
    i <- i + 1
    j <- j + 1
    } else {  
    if (cb1$clades[ii] < cb2$clades[jj]) {
      rfdmiss <- rfdmiss + cb1$brlens[ii]
      rfdtop <- rfdtop + 1
      i <- i + 1
      } else {
      rfdmiss <- rfdmiss + cb2$brlens[jj]
      rfdtop <- rfdtop + 1
      j <- j + 1        
      }
    }
  }
while (i <= len1) {
  ii <- ord1[i]
  rfdmiss <- rfdmiss + cb1$brlens[ii]
  rfdtop <- rfdtop + 1
  i <- i + 1 
  }
while (j <= len2) {
  jj <- ord2[j]
  rfdmiss <- rfdmiss + cb2$brlens[jj]
  rfdtop <- rfdtop + 1
  j <- j + 1 
  }
list(rfdmiss=rfdmiss, rfd=rfdmiss + rfdhit, rfdtop=rfdtop)
}

# Recursively makes a list of sequences of 1s and 2s.
# enum(character(0), NULL, 1, 3) makes a list with 8 numeric vectors
# 1,1,1; 1,1,2; ... 2,2,2
enum <- function(xxx, txt, n, m) {
if (n==m) {
  xxx <- c(xxx, list(c(txt, 1)))
  xxx <- c(xxx, list(c(txt, 2)))  
  } else {
  xxx <- enum(xxx, c(txt, 1), n+1, m)
  xxx <- enum(xxx, c(txt, 2), n+1, m)
  }
xxx
}


# Finds the Robinson-Foulds distance generalized for diploid-allotetraploid
# MUL-trees.
# MUL-tree 1 is passed as a clades and branch lengths.
# MUL-tree 2 is passed as a trlbi with no sequence info (no "A"s or "B"s i
# tip names) and a indices of the tips which are tetraploid
multree.rfd <- function(mt1.clades, mt2.trlbi, mt2.tetpairs) {
min.rfdi <- list(rfdmiss=Inf, rfd=Inf, rfdtop=Inf)
num <- length(mt2.tetpairs$tet1s)
xxx <- enum(character(0), NULL, 1, num)
for (i in 1:length(xxx)) {
  swaps <- xxx[[i]]
  tetpairs <- mt2.tetpairs
  for (j in 1:num) {
    if (swaps[j] == 2) {
      tmp <- tetpairs$tet1s[j]
      tetpairs$tet1s[j] <- tetpairs$tet2s[j]
      tetpairs$tet2s[j] <- tmp
      }    
    }
  mt2.clades <- trlbi.tetpairs.to.clades(mt2.trlbi, tetpairs)
  rfdi <- robinsonfouldsdist(mt1.clades, mt2.clades)
  if (rfdi$rfdmiss < min.rfdi$rfdmiss) {
    min.rfdi$rfdmiss <- rfdi$rfdmiss
    }
  if (rfdi$rfd < min.rfdi$rfd) {
    min.rfdi$rfd <- rfdi$rfd
    }
  if (rfdi$rfdtop < min.rfdi$rfdtop) {
    min.rfdi$rfdtop <- rfdi$rfdtop
    }
  }
min.rfdi 
}


#
#
#dpath <- "C:/Users/Graham/AAA/Programming/GoteborgSim/2012-10-31/J/Jg2n1t40/"
#fpath <- glue(dpath, "J5MTCg2n1t40r1.txt")
#             
#trueMtree.trlbi <- nodetable.to.trlbi(glue(dpath, "J1MTg2n1t40.txt"))
#trueMtree.trlbi <- remove.seqchars(trueMtree.trlbi, "A", "B")
#trueMtree.tetpairs <- find.tet.pairs(trueMtree.trlbi)
#trueMtree.trlbi <- add.seqchars(trueMtree.trlbi, trueMtree.tetpairs, "A", "B")
#trueMtree.clades <- trlbi.to.clades(trueMtree.trlbi)
#
#concensusMtree <- tree.string.from.nexus(glue(dpath, "J5MTCg2n1t40r1.txt"))               
#concensusMtree.trlbi.tetpairs <- newick.string.to.trlbi.and.tetpairs(concensusMtree, "0", "1")
#concensusMtree.trlbi <- concensusMtree.trlbi.tetpairs$trlbi
#concensusMtree.tetpairs <- concensusMtree.trlbi.tetpairs$tetpairs
#      
#multree.rfd(trueMtree.clades, concensusMtree.trlbi, concensusMtree.tetpairs)
#      
# 
#